#include <filesystem>
#include <iostream>

#include <utility>
#include <fggl/gfx/window.hpp>
#include <fggl/gfx/camera.hpp>

#include <fggl/gfx/compat.hpp>
#include <fggl/gfx/ogl/compat.hpp>

#include <fggl/data/procedural.hpp>
#include <fggl/data/storage.hpp>
#include <fggl/util/chrono.hpp>
#include <fggl/util/service.h>
#include <fggl/ecs3/ecs.hpp>

#include <fggl/debug/debug.h>
#include <imgui.h>
#include <fggl/scenes/Scene.h>

constexpr bool showNormals = false;

template <typename T> int sgn(T val) {
    return (T(0) < val) - (val < T(0));
}

// prototype of resource discovery
void discover(const std::filesystem::path& base) {

	std::vector< std::filesystem::path > contentPacks;
	for ( auto& item : std::filesystem::directory_iterator(base) ) {

		// content pack detection
		if ( std::filesystem::is_directory( item ) ) {
			auto manifest = item.path() / "manifest.yml";
			if ( std::filesystem::exists( manifest ) ) {
				contentPacks.push_back( item.path() );
			}
		}
		
	}

	// what did we find?
	std::cerr << "found pack(s): " << std::endl;
	for ( auto& pack : contentPacks ) {
		std::cerr << pack << std::endl;
	}

}

enum camera_type { cam_free, cam_arcball };
camera_type cam_mode = cam_free;

//TODO proper input system
using namespace fggl::input;
using InputManager = std::shared_ptr<fggl::input::Input>;

void process_arcball(fggl::ecs3::World& ecs, InputManager input, fggl::ecs::entity_t cam) {
	// see https://asliceofrendering.com/camera/2019/11/30/ArcballCamera/
	auto* camTransform = ecs.get<fggl::math::Transform>(cam);
	auto* camComp = ecs.get<fggl::gfx::Camera>(cam);
	auto& mouse = input->mouse;

	glm::vec4 position(camTransform->origin(), 1.0f);
	glm::vec4 pivot(camComp->target, 1.0f);
	glm::mat4 view = glm::lookAt( camTransform->origin(), camComp->target, camTransform->up() );
	glm::vec3 viewDir = -glm::transpose(view)[2];
	glm::vec3 rightDir = glm::transpose(view)[0];

	float deltaAngleX = ( 2 * M_PI ); 
	float deltaAngleY = ( M_PI );
	float xAngle = ( -mouse.axisDelta(fggl::input::MouseAxis::X) ) * deltaAngleX;
	float yAngle = ( -mouse.axisDelta(fggl::input::MouseAxis::Y) ) * deltaAngleY;

	auto cosAngle = glm::dot( viewDir, fggl::math::UP );
	if ( cosAngle * sgn(deltaAngleY) > 0.99f ) {
		deltaAngleY = 0;
	}

	// rotate the camera around the pivot on the first axis
	glm::mat4x4 rotationMatrixX(1.0f);
	rotationMatrixX = glm::rotate( rotationMatrixX, xAngle, fggl::math::UP );
	position = ( rotationMatrixX * ( position - pivot ) ) + pivot;

	// rotate the camera aroud the pivot on the second axis
	glm::mat4x4 rotationMatrixY(1.0f);
	rotationMatrixY = glm::rotate(rotationMatrixY, yAngle, rightDir );
	glm::vec3 finalPos = ( rotationMatrixY * ( position - pivot ) ) + pivot;

	camTransform->origin( finalPos );
}

constexpr float ROT_SPEED = 0.05f;
constexpr float PAN_SPEED = 0.05f;
constexpr glm::mat4 MAT_IDENTITY(1.0f);


void process_freecam(fggl::ecs3::World& ecs, InputManager input, fggl::ecs::entity_t cam) {
	float rotationValue = 0.0f;
	glm::vec3 translation(0.0f);

	auto& keyboard = input->keyboard;
	auto code_q = glfwGetKeyScancode(GLFW_KEY_Q);
	auto code_e = glfwGetKeyScancode(GLFW_KEY_E);
	auto code_w = glfwGetKeyScancode(GLFW_KEY_W);
	auto code_s = glfwGetKeyScancode(GLFW_KEY_S);
	auto code_d = glfwGetKeyScancode(GLFW_KEY_D);
	auto code_a = glfwGetKeyScancode(GLFW_KEY_A);


	// calulate rotation (user input)
	if ( keyboard.down( code_q ) ) {
		rotationValue = ROT_SPEED; 
	} else if ( keyboard.down(code_e) ) {
		rotationValue = -ROT_SPEED;
	}

	// calulate movement (user input)
	if ( keyboard.down(code_w) ) {
		translation -= fggl::math::RIGHT;
	}

	if ( keyboard.down(code_s) ) {
		translation += fggl::math::RIGHT;
	}

	if ( keyboard.down(code_d) ) {
		translation += fggl::math::FORWARD;
	}

	if ( keyboard.down(code_a) ) {
		translation -= fggl::math::FORWARD;
	}

	// apply rotation/movement
	auto camTransform = ecs.get<fggl::math::Transform>(cam);
	auto camComp = ecs.get<fggl::gfx::Camera>(cam);

	glm::vec4 position( camTransform->origin(), 1.0f );
	glm::vec4 pivot( camComp->target, 1.0f );

	// apply movement
	if ( translation != glm::vec3(0.0f) ) {
		const auto rotation = (position - pivot);
		const float angle = atan2( rotation.x, rotation.z );
		const auto rotationMat = glm::rotate( MAT_IDENTITY, angle, fggl::math::UP );

		auto deltaMove = (rotationMat * glm::vec4( translation, 1.0f )) * PAN_SPEED;
		deltaMove.w = 0.0f;

		position += deltaMove;
		pivot += deltaMove;
	}

	// apply rotation
	if ( rotationValue != 0.0f ) {
		glm::mat4 rotation = glm::rotate( MAT_IDENTITY, rotationValue, fggl::math::UP );
		position = ( rotation * ( position - pivot ) ) + pivot;
	}

	camTransform->origin( position );
	camComp->target = pivot;
}

void process_edgescroll(fggl::ecs3::World& ecs, InputManager input, fggl::ecs::entity_t cam) {
	glm::vec3 translation(0.0f);

	auto& mouse = input->mouse;

	// calulate movement (user input)
	if ( mouse.axis( MouseAxis::Y ) < 0.9f ) {
		translation -= fggl::math::RIGHT;
	}

	if ( mouse.axis( MouseAxis::Y) > -0.9f ) {
		translation += fggl::math::RIGHT;
	}

	if ( mouse.axis( MouseAxis::X) > -0.9f ) {
		translation += fggl::math::FORWARD;
	}

	if ( mouse.axis( MouseAxis::X ) < 0.9f ) {
		translation -= fggl::math::FORWARD;
	}

	// apply rotation/movement
	auto camTransform = ecs.get<fggl::math::Transform>(cam);
	auto camComp = ecs.get<fggl::gfx::Camera>(cam);

	glm::vec4 position( camTransform->origin(), 1.0f );
	glm::vec4 pivot( camComp->target, 1.0f );

	// apply movement
	if ( translation != glm::vec3(0.0f) ) {
		const auto rotation = (position - pivot);
		const float angle = atan2( rotation.x, rotation.z );
		const auto rotationMat = glm::rotate( MAT_IDENTITY, angle, fggl::math::UP );

		auto deltaMove = (rotationMat * glm::vec4( translation, 1.0f )) * PAN_SPEED;
		deltaMove.w = 0.0f;

		position += deltaMove;
		pivot += deltaMove;
	}

	// move camera
	camTransform->origin( position );
	camComp->target = pivot;
}


void process_camera(fggl::ecs3::World& ecs, InputManager input) {
    auto cameras = ecs.findMatching<fggl::gfx::Camera>();
    fggl::ecs3::entity_t cam = cameras[0];

	auto camTransform = ecs.get<fggl::math::Transform>(cam);
	auto camComp = ecs.get<fggl::gfx::Camera>(cam);

	const glm::vec3 dir = ( camTransform->origin() - camComp->target );
	const glm::vec3 forward = glm::normalize( dir );

	// scroll wheel
	glm::vec3 motion(0.0f);
	float delta = input->mouse.axis( fggl::input::MouseAxis::SCROLL_Y );
	if ( (glm::length( dir ) < 25.0f && delta < 0.0f) || (glm::length( dir ) > 2.5f && delta > 0.0f) )
		motion -= (forward * delta);

	camTransform->origin( camTransform->origin() + motion );

	if ( cam_mode == cam_arcball || input->mouse.button( fggl::input::MouseButton::MIDDLE ) ) { 
		process_arcball(ecs, input, cam);
	} else if ( cam_mode == cam_free ) {
		process_freecam(ecs, input, cam);
	}
	process_edgescroll( ecs, input, cam );
}


class MenuScene : public fggl::scenes::Scene {

public:
    explicit MenuScene(InputManager inputs) : m_inputs(std::move(inputs)) {}
    ~MenuScene() override = default;

    void setup() override {
    }

    void cleanup() override {

    }

    void update() override {
        bool leftMouse = m_inputs->mouse.button(fggl::input::MouseButton::LEFT);
        if (leftMouse) {
            auto scenes = fggl::util::ServiceLocator::instance().providePtr<fggl::scenes::SceneManager>();
            scenes->activate("game");
        }
    }

    void render() override {

    }

    private:
        InputManager m_inputs;

};

class GameScene : public fggl::scenes::Scene {
public:
    explicit GameScene(fggl::ecs3::World& world, InputManager inputs) : m_world(world), m_inputs(std::move(inputs)) { };
    ~GameScene() override = default;

    void setup() override {
        auto types = m_world.types();

        // create camera using strings
        {
            auto prototype = m_world.create(false);
            m_world.add(prototype, types.find(fggl::math::Transform::name));
            m_world.add(prototype, types.find(fggl::gfx::Camera::name));

            auto camTf = m_world.get<fggl::math::Transform>(prototype);
            camTf->origin( glm::vec3(0.0f, 3.0f, 3.0f) );
        }

        // create building prototype
        fggl::ecs3::entity_t bunker;
        {
            bunker = m_world.create(true);
            m_world.add(bunker, types.find(fggl::math::Transform::name));

            // mesh
            int nSections = 2;
            constexpr float HALF_PI = M_PI / 2.0f;
            constexpr char shader[] = "phong";

            fggl::data::Mesh mesh;
            for (int j=-(nSections/2); j<=nSections/2; j++) {
                const auto shapeOffset = glm::vec3( 0.0f, 0.0f, j * 1.0f );

                const auto cubeMat = glm::translate( fggl::math::mat4( 1.0f ) , shapeOffset );
                const auto leftSlope = fggl::math::modelMatrix(
                        glm::vec3(-1.0f, 0.0f, 0.0f) + shapeOffset,
                        glm::vec3( 0.0f, -HALF_PI, 0.0f) );
                const auto rightSlope = fggl::math::modelMatrix(
                        glm::vec3( 1.0f, 0.0f, 0.0f) + shapeOffset,
                        glm::vec3( 0.0f, HALF_PI, 0.0f) );

                fggl::data::make_cube( mesh, cubeMat );
                fggl::data::make_slope( mesh, leftSlope );
                fggl::data::make_slope( mesh, rightSlope );
            }
            mesh.removeDups();

            fggl::gfx::StaticMesh staticMesh{mesh, shader};
            m_world.set<fggl::gfx::StaticMesh>(bunker, &staticMesh);
        }

        int nCubes = 3;
        for ( int i=0; i<nCubes; i++ ) {
            auto bunkerClone = m_world.copy(bunker);
            auto result = m_world.get<fggl::math::Transform>(bunkerClone);
            result->origin( glm::vec3( i * 5.0f, 0.0f, 0.0f) );
        }
    }

    void cleanup() override {

    }

    void update() override {
        process_camera(m_world, m_inputs);
    }

    void render() override {
        auto types = m_world.types();

        // ECS3 inspector
        ImGui::Begin("Entities");
        auto entityItr = m_world.all();
        for (auto& entity : entityItr) {
            std::string label = "entity-" + std::to_string(entity);
            if ( ImGui::TreeNode(label.c_str()) ){
                auto entComp = m_world.getComponents(entity);
                for ( auto comp : entComp ){
                    auto meta = types.meta(comp);
                    ImGui::Text("%s (%d) - %lu bytes", meta->name(), comp, meta->size());
                }
                ImGui::TreePop();
            }
        }
        ImGui::End();

        // imgui gamepad debug
        if ( m_gamepadWindow ) {
            auto &gamepads = m_inputs->gamepads;
            ImGui::Begin("GamePad", &m_gamepadWindow);
            for (int i = 0; i < 16; i++) {
                std::string title = gamepads.name(i);

                bool present = gamepads.present(i);
                if (ImGui::TreeNode(title.c_str())) {
                    ImGui::Text("present: %s", present ? "yes" : "no");

                    if (present) {

                        if (ImGui::TreeNode("buttons##2")) {
                            for (auto &btn: fggl::input::GamepadButtonsMicrosoft) {
                                ImGui::Text("%s: %i %i %i", btn.name,
                                            gamepads.button(i, btn.id),
                                            gamepads.buttonPressed(i, btn.id),
                                            gamepads.buttonReleased(i, btn.id)
                                );
                            }
                            ImGui::TreePop();
                        }

                        if (ImGui::TreeNode("axes##2")) {
                            for (auto &axis: fggl::input::GamepadAxes) {
                                ImGui::Text("%s: %f %f", axis.name,
                                            gamepads.axis(i, axis.id),
                                            gamepads.axisDelta(i, axis.id)
                                );

                            }
                            ImGui::TreePop();
                        }

                    }

                    ImGui::TreePop();
                    ImGui::Separator();
                }

            }
            ImGui::End();
        }
    }

private:
    fggl::ecs3::World& m_world;
    InputManager m_inputs;
    bool m_gamepadWindow = false;
};

int main(int argc, char* argv[]) {
    auto& locator = fggl::util::ServiceLocator::instance();

    // setup ECS
    fggl::ecs3::TypeRegistry types;
    types.make<fggl::math::Transform>();
    fggl::ecs3::ModuleManager modules(types);

    fggl::ecs3::World ecs(types);

    // input management
    auto inputs = std::make_shared<fggl::input::Input>();
    locator.supply<fggl::input::Input>(inputs);

    auto glfwModule = modules.load<fggl::gfx::ecsGlfwModule>(inputs);

    // window
    fggl::gfx::Window win{};
    win.title("FGGL Demo");
	win.fullscreen( true );

	// storage API
	fggl::data::Storage storage;
	//discover( storage.resolvePath(fggl::data::Data, "../../packs") );

	// Opengl APIs
    auto glModule = modules.load<fggl::gfx::ecsOpenGLModule>(win, storage);

	fggl::gfx::loadPipeline(glModule, "unlit", false);
	fggl::gfx::loadPipeline(glModule, "phong", false);
	fggl::gfx::loadPipeline(glModule, "normals", false);

	// debug layer
	fggl::debug::DebugUI debug(win);
	debug.visible(true);

    // Create scene manager
    auto scenes = std::make_shared<fggl::scenes::SceneManager>();
    locator.supply<fggl::scenes::SceneManager>(scenes);

    scenes->create("main_menu", std::make_shared<MenuScene>(inputs));
    scenes->create("game", std::make_shared<GameScene>(ecs, inputs));

    // activate the main menu
    scenes->activate("main_menu");

	// create ECS

	fggl::util::Timer time{};
	time.frequency( glfwGetTimerFrequency() );
	time.setup( glfwGetTimerValue() );

	while( !win.closeRequested() ) {
		//
		// Setup setup
		//
		time.tick( glfwGetTimerValue() );
		inputs->frame( time.delta() );

		glfwModule->context.pollEvents();
		debug.frameStart();

		//
		// update step
		//
        scenes->update();

        // render the scene
        win.activate();
        glModule->ogl.clear();

        // allow the scene to do stuff, then actually render
        scenes->render();
        fggl::gfx::renderMeshes(glModule, ecs, time.delta());

        debug.draw();
        win.swap();
    }

	return 0;
}
